"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.IdentityManager = void 0;
const axios_1 = __importDefault(require("axios"));
const crypto_1 = require("crypto");
const https_1 = __importDefault(require("https"));
const apiError_1 = require("../model/apiError");
const ProjectValidation = __importStar(require("../service/projectValidation"));
const backtraceUniverseHelper_1 = require("../universe/backtraceUniverseHelper");
/**
 * Identity manager
 */
class IdentityManager {
    constructor(descr, logger) {
        this.descr = descr;
        /*
         * Stores opaque objects per authenticated remote coronerd.
         * Available for consumers to use as they see fit.
         */
        this.authenticated = {};
        /*
         * Stores coronerd values at the register step,
         * to be attached during authRequest step.
         * Uses universe names as key.
         */
        this.registeredUniverses = {};
        this.httpsAgent = this.descr.isOnPremise
            ? new https_1.default.Agent({ rejectUnauthorized: false })
            : https_1.default.globalAgent;
        if (logger) {
            this.setLogger(logger);
        }
    }
    /**
     * Sets the coroned callback function - the funciton that will be
     * executed by the identity manager, when a new coronerd instance
     * registers. Setting a new coronerd callback will also invoke the
     * callback funciton with all saved unique ip addresses.
     * @param callback: callback function that retrieves a registered coronerd ip address
     */
    coronerdCallback(callback) {
        this.onCoronerCallback = callback;
        const authenticatedIpAddresses = Object.keys(this.authenticated);
        authenticatedIpAddresses.forEach(async (ip) => {
            /**
             * Pass false as a alreadyRegisteredFlag to tell
             * the callback function, this is a new coroned address for
             * the callback consumer.
             */
            await callback(ip, false);
        });
    }
    setLogger(logger) {
        if (!logger || !logger.log) {
            throw new TypeError('unusable logger passed in');
        }
        this.logger = logger;
    }
    /*
     * Generate a middleware handler for Express.js, using options to hook the
     * service up.
     *
     * @param opts ICoronerRequestOption - coronerd url callback if desired
     */
    serviceRequest(opts) {
        if (opts?.coronerdCallback) {
            this.onCoronerCallback = opts.coronerdCallback;
        }
        /*
         * Handle <prefix>/service requests.  These require action QS.  Other QS
         * are defined per action.
         */
        return [
            async (request, response, next) => {
                if (!request.body) {
                    return next(new apiError_1.ApiError('Missing body', 400));
                }
                const action = request.body.action;
                if (!action) {
                    return next(new apiError_1.ApiError('Missing action parameter', 400));
                }
                switch (action) {
                    case 'register':
                        const { url, nonce, hmac, service_to_coronerd_url } = request.body;
                        if (!url || !nonce || !hmac) {
                            return next(new apiError_1.ApiError(`missing parameters
                  ${!url ? 'url' : ''}
                  ${!nonce ? 'nonce' : ''}
                  ${!hmac ? 'hmac' : ''}`, 400));
                        }
                        if (!this.checkHmac(this.descr.secret, nonce, hmac)) {
                            return next(new apiError_1.ApiError('Unauthorized', 403));
                        }
                        /* Track authenticated URLs. */
                        const alreadyRegistered = !!this.authenticated[url];
                        if (!alreadyRegistered) {
                            this.authenticated[url] = {};
                        }
                        /*
                         * Track registered coronerds information so that it
                         *  can be passed to consumer through authenticateRequest's
                         *  coronerAuth object. This asynchronously updates
                         *  local store with coronerdData for each universe
                         *  discovered in a connected coronerd.
                         */
                        this.retrieveAndStoreCoronerdInfo({
                            coronerdUrl: url,
                            coronerdData: { service_to_coronerd_url },
                            hmac,
                            nonce,
                        });
                        if (this.onCoronerCallback) {
                            await this.onCoronerCallback(url, alreadyRegistered);
                        }
                        // Ignore project validation in service layer handshake.
                        ProjectValidation.setValid(response);
                        response.json({
                            status: 'ok',
                        });
                        break;
                    default:
                        throw new apiError_1.ApiError('invalid action', 400);
                }
            },
        ];
    }
    /**
     * Generate a middleware for Express.js which validates a request
     * by comparing universe/project data avaialble in the request object
     * with universe/project data available in the Authenticated request.
     *
     * This middleware must be used after the authenticateRequest.
     * @param dataProvider universe/project provider.
     */
    validateProjectAccess(dataProvider) {
        return (req, res, next) => {
            const authRequest = req;
            if (!authRequest.coronerAuth || !authRequest.coronerAuth.cfg) {
                return next(new apiError_1.ApiError('Cannot verify the user access to specific project/universe.', 500));
            }
            const { universe, project } = dataProvider(authRequest);
            if (!(universe in authRequest.coronerAuth.cfg.universes)) {
                return next(new apiError_1.ApiError('Unauthorized.', 401));
            }
            if (project && !this.checkProjectAccess(req, res, [project])) {
                return next(new apiError_1.ApiError('Unauthorized.', 401));
            }
            this.bypassProjectAccessCheck(res);
            next();
        };
    }
    /**
     * Generate a middleware for Express.js, which authenticates a request given
     * its URL and session token, provided by an earlier callback in
     * req.coronerAuth.
     * Upon success, the configuration data will be in req.coronerAuth.cfg.
     */
    authenticateRequest() {
        return async (request, _, next) => {
            const { url, token } = this.readCoronerAuthInfo(request);
            if (!token || !url) {
                return next(new apiError_1.ApiError(`${request.ip}: Missing internal params = token or url`, 403));
            }
            const authRequest = request;
            if (!authRequest) {
                return next(new apiError_1.ApiError(`Invalid request cast`, 500));
            }
            try {
                const serverConfiguration = await this.authenticate(url, token);
                if (!serverConfiguration || !serverConfiguration.user) {
                    return next(new apiError_1.ApiError('Auth failure', 403));
                }
                /**
                 * Append service_to_coronerd_url if available.
                 *   It comes from coronerds that have successfully registered
                 *   with the service and have cached their info by universeName in registeredUniverese.
                 */
                const requestedUniverse = serverConfiguration.universe.name;
                const registeredUniverse = this.registeredUniverses[requestedUniverse];
                if (!registeredUniverse) {
                    this.log('warn', `universe ${requestedUniverse} not found in registeredUniverses`);
                }
                authRequest.coronerAuth = {
                    url,
                    token,
                    cfg: serverConfiguration,
                    service_to_coronerd_url: registeredUniverse?.service_to_coronerd_url,
                };
                next();
            }
            catch (err) {
                if (axios_1.default.isAxiosError(err) && err.response?.status) {
                    next(new apiError_1.ApiError(err.message, err.response.status));
                }
                else {
                    next(err);
                }
            }
        };
    }
    /**
     * This assumes route has already passed authentication through authenticateRequest
     * and uses /api/config result from that to check projects.
     */
    checkProjectAccess(request, response, projects) {
        const authRequest = request;
        const universe = authRequest?.coronerAuth?.cfg?.universe;
        const projectsExt = universe &&
            authRequest?.coronerAuth?.cfg?.universes[universe.name]?.projects_ext;
        if (!projectsExt) {
            this.log('debug', `missing projects_ext`);
            return false; // invalidated because flag is never set to true
        }
        return ProjectValidation.check(response, projectsExt, projects);
    }
    /**
     * Unconditionally set projectValidated flag to true.
     * Allows service routes that do not need project validation to procceed with
     * typical Express send method.
     */
    bypassProjectAccessCheck(response) {
        ProjectValidation.setValid(response);
    }
    /**
     * Use coronerd auth method to validate url and session token provided by user
     * For Express apps, consider using the authenticateRequest middleware instead
     * @param url coronerd url
     * @param token coronerd token
     */
    async authenticate(url, token) {
        if (!token || !url) {
            throw new apiError_1.ApiError('missing url/token', 400);
        }
        const prefix = url.endsWith('/') ? '' : '/';
        const result = await axios_1.default.get(`${url}${prefix}api/config`, {
            httpsAgent: this.httpsAgent,
            headers: {
                'X-Service-Name': this.descr.name,
                'X-Coroner-Token': this.sanitizeHeader(token),
                'X-Service-HMAC': this.generateHmac(this.descr.secret, token),
            },
        });
        if (result.status !== 200) {
            throw new apiError_1.ApiError(`Invalid server response code ${result.status}`, result.status);
        }
        /* Verify the coronerd knows the service secret. */
        const coronerd_nonce = result.headers['x-service-nonce'];
        const coronerd_hmac = result.headers['x-service-hmac'];
        if (!coronerd_nonce || !coronerd_hmac) {
            throw new apiError_1.ApiError('Missing response parameters - coronerd_nonce/hmac', 400);
        }
        if (!this.checkHmac(this.descr.secret, coronerd_nonce, coronerd_hmac)) {
            throw new apiError_1.ApiError('Invalid server generated HMAC', 400);
        }
        return result.data;
    }
    /**
     * Login to a coronerd.
     * @param Coronerd base URL to login to
     */
    async loginCoronerd(coronerdUrl) {
        if (!coronerdUrl) {
            throw new Error('CoronerdUrl is undefined or empty string');
        }
        const prefix = coronerdUrl.endsWith('/') ? '' : '/';
        const protocol = coronerdUrl.startsWith('http') ? '' : backtraceUniverseHelper_1.http;
        const requestUrl = `${protocol}${coronerdUrl}${prefix}api/login`;
        const nonce = (0, crypto_1.randomBytes)(32).toString('hex');
        const hmac = this.generateHmac(this.descr.secret, nonce);
        try {
            const result = await axios_1.default.post(requestUrl, null, {
                httpsAgent: this.httpsAgent,
                headers: {
                    'X-Service-Name': this.descr.name,
                    'X-Service-Nonce': nonce,
                    'X-Service-HMAC': hmac,
                },
            });
            return result.status === 200 ? result.data : undefined;
        }
        catch (err) {
            this.logger?.error(err, `Cannot login to coronerd. Requested address: ${requestUrl}`);
            return undefined;
        }
    }
    /**
     * Retrieves the zeroconf from a coronerd
     * to store data for later client consumption.
     */
    async retrieveAndStoreCoronerdInfo({ coronerdUrl, coronerdData, hmac, nonce, }) {
        if (!coronerdUrl) {
            throw new Error('CoronerdUrl is undefined or empty string');
        }
        const prefix = coronerdUrl.endsWith('/') ? '' : '/';
        const protocol = coronerdUrl.startsWith('http') ? '' : backtraceUniverseHelper_1.http;
        try {
            const bpgUniverseRequestUrl = `${protocol}${coronerdUrl}${prefix}api/bpg`;
            const bpgUniverseRequestData = {
                actions: [{ action: 'get', type: 'configuration/universe' }],
            };
            const bpgUniverseResult = await axios_1.default.post(bpgUniverseRequestUrl, bpgUniverseRequestData, {
                httpsAgent: this.httpsAgent,
                headers: {
                    'X-Service-Name': this.descr.name,
                    'X-Service-Nonce': nonce,
                    'X-Service-HMAC': hmac,
                    'Content-Type': 'application/json',
                },
            });
            // select the single result for our single BPG action request
            const bpgUniverseObjects = bpgUniverseResult.data.results[0].result;
            bpgUniverseObjects.forEach((universe) => {
                if (universe.deleted !== 1) {
                    this.registeredUniverses[universe.name] = {
                        ...coronerdData,
                    };
                }
            });
        }
        catch (err) {
            this.logger?.error(err, 'Cannot get universes from coronerd.');
            return undefined;
        }
    }
    /**
     * @deprecated: these methods will be removed in new backtrace-service  update
     * Check if authtoken is a valid token generated by coronerd
     * @param universeUrl url to coronerd universe
     * @param authToken user token
     */
    async isValidToken(universeUrl, authToken) {
        const result = await this.getConfiguration(universeUrl, authToken);
        return result !== undefined;
    }
    /**
     * @deprecated: these methods will be removed in new backtrace-service  update
     * Get Backtrace configuration
     * @param universeUrl url to coronerd universe
     * @param authToken user token
     */
    async getConfiguration(universeUrl, authToken) {
        if (!authToken || !universeUrl) {
            return undefined;
        }
        const prefix = universeUrl.endsWith('/') ? '' : '/';
        const protocol = universeUrl.startsWith('http') ? '' : backtraceUniverseHelper_1.http;
        const requestUrl = `${protocol}${universeUrl}${prefix}api/config`;
        try {
            const result = await axios_1.default.get(requestUrl, {
                headers: {
                    'X-Coroner-Token': authToken,
                },
            });
            return result.status === 200 ? result.data : undefined;
        }
        catch (err) {
            return undefined;
        }
    }
    generateHmac(secret, nonce) {
        if (!secret || !nonce) {
            return '';
        }
        return (0, crypto_1.createHmac)('sha256', secret).update(nonce).digest('hex');
    }
    checkHmac(secret, nonce, proposedHMac) {
        const hmac = this.generateHmac(secret, nonce);
        if (!hmac) {
            return false;
        }
        return hmac === proposedHMac;
    }
    sanitizeHeader(value) {
        if (!value) {
            return '';
        }
        return value.replace(/[^a-z0-9A-Z?!]/gi, '');
    }
    log(level, msg) {
        if (this.logger) {
            this.logger.log(level, msg);
        }
    }
    readCoronerAuthInfo(request) {
        const authRequest = request;
        let data = authRequest.coronerAuth;
        if (!data) {
            data = {
                token: this.sanitizeHeader(request.get('X-Coroner-Token')),
                url: request.get('X-Coroner-Location') ||
                    request.get('origin'),
            };
            if (data.url && !data.url.startsWith('http')) {
                data.url = `${backtraceUniverseHelper_1.http}${data.url}`;
            }
        }
        return data;
    }
}
exports.IdentityManager = IdentityManager;
// vim:set ts=2 sw=2 et:
//# sourceMappingURL=identity.js.map