"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MetricsStorage = void 0;
const axios_1 = __importDefault(require("axios"));
const os_1 = __importDefault(require("os"));
const logger_1 = require("../log/logger");
class MetricsStorage {
    constructor(uuid, secret, opts, app, logger = logger_1.Logger.getLogger()) {
        this.uuid = uuid;
        this.secret = secret;
        this.opts = opts;
        this.logger = logger;
        this._hostUrl = 'https://api.circonus.com/module/httptrap/';
        this._url = `${this._hostUrl}${this.uuid}/${this.secret}`;
        this._defaultMetricsInterval = 60000;
        this._hostName = os_1.default.hostname().split('.')[0];
        this._data = {};
        if (!this.uuid || !this.secret) {
            if (logger) {
                logger.warn('Opts (or uuid/secret) object is required. Metrics are disabled');
            }
            return;
        }
        this.logger.debug('Initializing circonus metrics integrations');
        // override default circonus url.
        // different url might be useful in case
        // when we will prepare our own metrics client
        if (this.opts && this.opts.hostUrl) {
            this._hostUrl = this.opts.hostUrl;
        }
        if (app && opts?.disableExpressMetrics !== true) {
            app.use((req, res, next) => this.registerRequestLatency(req, res, next));
        }
        this.addMetric('birthday', Date.now());
        setInterval(async () => {
            if (this.beforeSend) {
                await this.beforeSend();
            }
            if (Object.keys(this._data).length > 0) {
                this.send();
            }
        }, (this.opts && this.opts.interval) || this._defaultMetricsInterval);
    }
    /**
     * Convert objects to metrics
     * @param source metrics object
     */
    add(source) {
        // metrics are disabled because of missing secret and uuid parameters
        if (!this.uuid || !this.secret) {
            return;
        }
        for (const key in source) {
            if (source.hasOwnProperty(key)) {
                const value = source[key];
                this.addMetric(key, value);
            }
        }
    }
    /**
     * Add single value to metrics
     * @param name metrics name
     * @param value metrics value
     */
    addMetric(name, value) {
        // metrics are disabled because of missing secret and uuid parameters
        if (!this.uuid || !this.secret) {
            return;
        }
        name = `${name}|ST[hostname:${this._hostName}]`;
        this._data[name] = this._data[name] || { _type: 'n', _value: [] };
        this._data[name]._value.push(value);
    }
    /**
     * Add single value to a histogram type metric
     * This is the same as addMetric except for _type: 'h'
     * @param name metrics name
     * @param value metrics value
     */
    addMetricHistogram(name, value) {
        // metrics are disabled because of missing secret and uuid parameters
        if (!this.uuid || !this.secret) {
            return;
        }
        name = `${name}|ST[hostname:${this._hostName}]`;
        this._data[name] = this._data[name] || { _type: 'h', _value: [] };
        this._data[name]._value.push(value);
    }
    registerRequestLatency(req, res, next) {
        if (!req.path) {
            return;
        }
        const startHrTime = process.hrtime();
        res.on('finish', () => {
            const elapsedHrTime = process.hrtime(startHrTime);
            const elapsedTimeInMs = elapsedHrTime[0] * 1000 + elapsedHrTime[1] / 1e6;
            this.addMetric('express`' + req.method + '`' + req.path + '`latency', elapsedTimeInMs);
        });
        next();
    }
    async send() {
        const data = this._data;
        this._data = {};
        try {
            await axios_1.default.put(this._url, data);
        }
        catch (error) {
            this.logger?.warn(error, 'Cannot send metrics.');
        }
    }
}
exports.MetricsStorage = MetricsStorage;
//# sourceMappingURL=metricsStorage.js.map