"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.BacktraceService = void 0;
const argparse_1 = require("argparse");
const body_parser_1 = require("body-parser");
const compression_1 = __importDefault(require("compression"));
const cors_1 = __importDefault(require("cors"));
const express_1 = __importDefault(require("express"));
const helmet_1 = __importDefault(require("helmet"));
const nocache_1 = __importDefault(require("nocache"));
const __1 = require("..");
const logger_1 = require("../log/logger");
const metricsStorage_1 = require("../metrics/metricsStorage");
const errorScope_1 = require("../model/errorScope");
const projectValidation_1 = require("./projectValidation");
const serviceConfiguration_1 = require("./serviceConfiguration");
// tslint:disable-next-line: no-var-requires
const packageJson = require('../../package.json');
class BacktraceService {
    constructor(name, port, _opts) {
        this.name = name;
        this.port = port;
        this._opts = _opts;
        this.coronerdAddress = new Set();
        this.app = _opts?.app ? _opts.app : (0, express_1.default)();
        this.initializeDefaultArgParser(this._opts?.serviceMutation);
        this.configuration = BacktraceService.readConfiguration(this.name, this.port, this.arguments?.configuration || this._opts?.confPath, false);
        // the descriptor field is always available
        // since it can't be igored by the readconfiguration parameters
        // if the descriptor is not available, then the service refuses
        // to start
        this._descriptor = this.configuration.descriptor;
        this.apiPrefix = this._descriptor.resource;
        this._logger = this.initializeLogger();
        /**
         * In previous version of backtrace-service we assumed that MetricsStorage
         * will always exists. To make sure metricsStorage object is always available
         * for services that already use this object from BacktraceService
         * we will pass empty uuid/secret if remoteLog is disabled.
         */
        const { uuid, secret } = this.getMetricsSecret();
        this.metrics = new metricsStorage_1.MetricsStorage(uuid, secret, {
            interval: this.configuration?.metrics?.interval,
            hostUrl: this.configuration?.metrics?.hostUrl,
            disableExpressMetrics: this.configuration?.metrics
                ?.disableExpressMetrics,
        }, this.app, this._logger);
        this.identityManager = new __1.IdentityManager(this._descriptor, this._logger);
        BacktraceService._service = this;
    }
    /**
     * Returns ArgumentParser instance arguments.
     */
    get arguments() {
        if (!this._arguments) {
            this._arguments = this._argumentParser.parseKnownArgs()[0];
        }
        return this._arguments;
    }
    /**
     * Create and startup express application with basic options.
     * By using this method, application will apply all
     * Backtrace service rules
     * @param name service name (required by service layer)
     * @param port service port (required by service layer). Note this
     * service requested port. Service layer might assign a different
     * service port based on the service descriptor.
     * @param opts optional application options
     * @param expressApp - deprecated. Optional express app instance.
     * Please use expressInstance from app instead.
     * By using this parameter you can create express app in different
     * framework and still be able to use Backtrace integrations
     */
    static bootstrap(name, port, opts, app) {
        if (!port || !name) {
            throw new Error(`Invalid arguments. Name: ${name}, port: ${port}`);
        }
        // backward compatibility
        if (app && opts) {
            opts.app = app;
        }
        const service = new BacktraceService(name, port, opts);
        // assign express intsance prepared by backtrace service/passed by user
        const expressApp = service.app;
        this.setupErrorScopeMiddleware(expressApp, opts?.errorScopeOptions);
        this.setupAuth(expressApp);
        this.setupMiddlewares(expressApp, service.apiPrefix, service.identityManager);
        this.setupServiceLayer(expressApp, name, port, (ip) => {
            service.coronerdAddress.add(ip);
            if (opts && opts.coronerdCallback) {
                opts.coronerdCallback(ip);
            }
        });
        process.title = name;
        expressApp.on('close', () => {
            service._logger.warn('server is closing');
        });
        return service;
    }
    /**
     * Read and parse service configuration
     * @param name service name
     * @param port service port number
     * @param path (optional) service configuration path
     * @param ignoreDescriptor (optional) determine if service
     * should create/read service layer service descriptor
     * @param parser argParser instance to read command line arguments
     */
    static readConfiguration(name, port, path, ignoreDescriptor) {
        const serviceConfiguration = new serviceConfiguration_1.ServiceConfiguration();
        const possiblePaths = path ? [path] : [];
        possiblePaths.push(...serviceConfiguration.getProjectSettingsPaths(name));
        const configuration = serviceConfiguration.loadConfiguration(possiblePaths);
        if (!ignoreDescriptor) {
            configuration.descriptor = (0, __1.getDescriptor)(name, port);
        }
        return configuration;
    }
    /**
     * Setup popular Backtrace middlewares
     * @param app express app
     */
    static setupMiddlewares(app, prefix, identityManager) {
        app.use((0, body_parser_1.json)({
            verify: (res, _, buf) => {
                res.rawBody = buf;
            },
        }));
        app.use((0, body_parser_1.urlencoded)({
            extended: true,
            verify: (res, _, buf) => {
                res.rawBody = buf;
            },
        }));
        app.use((0, compression_1.default)());
        // Add default Backtrace headers to every service response
        app.use((req, res, next) => {
            // add service version
            res.set('service-version', process.env.npm_package_version);
            // add backtrace-service version
            res.set('bt-service-version', packageJson.version);
            next();
        });
        // Route hit by pingdom
        app.get(prefix ? prefix : '/', (_req, res) => {
            identityManager?.bypassProjectAccessCheck(res);
            res.type('xml');
            res.status(200).send(`<html></html>`);
        });
    }
    static setupErrorScopeMiddleware(app, opts) {
        app.use((req, res, next) => {
            const errorScope = new errorScope_1.ErrorScope();
            if (opts?.addExpressAttributes) {
                errorScope.addAttributes(() => ({
                    originalUrl: req.originalUrl,
                    protocol: req.protocol,
                    query: req.query,
                    headers: req.headers,
                    body: { value: req.body }, // Avoid putting string bodies in attributes
                }));
            }
            req.errorScope = errorScope;
            next();
        });
    }
    /**
     * Setup express error middleware
     * @param app express app
     */
    static setupErrorMiddleware(app, identityManager, overrideLogger) {
        app.use((err, req, res, next) => {
            // handle a situation when we already sent a response from onError method
            // we shouldn't do that and we should let error middleware handle that
            // but here we wan't to make sure our code won't break up
            // and we won't generate another error
            if (res.headersSent) {
                return;
            }
            const logger = overrideLogger ?? logger_1.Logger.getLogger();
            // if we reach this part of code, the error middlewa needs to assume
            // the teams validation weren't handled on the route side - for example:
            // when exception were throwed by the 3rd party library when someone sends
            // invalid JSON object. In this situation we should return an error and mark
            // request as validated
            identityManager?.bypassProjectAccessCheck(res);
            // Backtrace service ApiError contains type statusCode that allows us
            // to determine what type of response we should return
            // in addition to that the error middleware can read a status code from response.status
            // property. This is a popular pattern for axios communication - check other services
            // to learn more how errors throwed by axios are beings handled by them.
            const { errorScope } = req;
            if (err.statusCode || (err.response && err.response.status)) {
                const status = err.statusCode || err.response.status;
                const message = err.message || err.response.message;
                if (status === 500) {
                    const attributes = errorScope?.getAttributes();
                    if (attributes) {
                        logger.error(err, attributes);
                    }
                    else {
                        logger.error(err);
                    }
                }
                else {
                    logger.warn(message);
                }
                res.status(status).send(message);
            }
            else {
                const attributes = errorScope?.getAttributes();
                if (attributes) {
                    logger.error(err, attributes);
                }
                else {
                    logger.error(err);
                }
                res.status(500).send({
                    error: 'Internal server error',
                });
            }
        });
        app.use((req, res, next) => {
            // prevent overriding headers.
            if (res.headersSent) {
                return;
            }
            identityManager?.bypassProjectAccessCheck(res);
            // response json might be required by front-end
            res.status(404).send({
                error: 'Not found',
            });
        });
    }
    /**
     * Setup security middlewares
     * @param app express app
     */
    static setupAuth(app) {
        const corsSettings = (0, cors_1.default)({
            exposedHeaders: ['service-version', 'bt-service-version'],
        });
        app.use(corsSettings);
        app.options('*', corsSettings);
        app.use((0, nocache_1.default)());
        app.use((0, helmet_1.default)({
            hsts: false,
        }));
        app.use(projectValidation_1.projectValidation);
    }
    /**
     * Setup service layer integration
     * @param app express app
     */
    static setupServiceLayer(app, name, port, coronerdCallback) {
        let serviceDescriptor = this._service
            ? this._service._descriptor
            : undefined;
        // this case handle a situation when we didn't call backtraceService constr
        // here we want to read service descriptor from hard drive - if descr exists
        // we're using here a service name and port instead of service descriptor
        // to avoid extra steps required by service to setup service layer.
        if (!serviceDescriptor && name && port) {
            serviceDescriptor = (0, __1.getDescriptor)(name, port);
        }
        if (!serviceDescriptor) {
            throw new Error('Descriptor is undefined');
        }
        const identityManager = this._service
            ? this._service.identityManager
            : new __1.IdentityManager(serviceDescriptor);
        app.post(`${serviceDescriptor.resource}/service`, identityManager.serviceRequest({
            name: serviceDescriptor.name,
            secret: serviceDescriptor.secret,
            coronerdCallback,
        }));
    }
    /**
     * Returns the information if the service is available in the on-premise environment
     * @returns true if the environment is set to on-premise
     */
    isOnPremise() {
        return this._descriptor.isOnPremise;
    }
    start(clusterSupport = false) {
        if (!this.app) {
            this._logger.error(`Express application is not defined. Cannot start application`);
            return;
        }
        if (!this._descriptor) {
            this._logger.error(`Descriptor object is not defined.
        Please execute setupServiceLayer method`);
            return;
        }
        if (this.port !== this._descriptor.port) {
            this._logger.warn(`Requested application port is different than service descriptor port.` +
                `Continuing using service descriptor port: ${this._descriptor.port}`);
        }
        BacktraceService.setupErrorMiddleware(this.app, this.identityManager, this._opts?.logger);
        (0, __1.listenDescriptor)(this._descriptor, this.app, this.configuration.ssl, () => {
            this._logger.info(`Application is Running. ${this._descriptor
                ? `Port: ${this._descriptor.port}. Api prefix: ${this._descriptor.resource}`
                : ''}`);
        }, clusterSupport);
    }
    /**
     * Initialize the service logger.
     * @returns Service logger instance.
     */
    initializeLogger() {
        const logLevel = this.configuration.logger && this.configuration.logger.logLevel
            ? this.configuration.logger.logLevel
            : undefined;
        const optionsLogger = this._opts && this._opts.logger;
        let logger;
        if (optionsLogger) {
            logger_1.Logger.setLogger(this._opts.logger);
            logger = this._opts.logger;
        }
        if (!logger) {
            logger = new logger_1.Logger(logLevel);
        }
        if (this.isBacktraceIntegrationEnabled()) {
            logger_1.Logger.initializeBacktrace(this.configuration.backtrace);
        }
        return logger;
    }
    /**
     * Initialize default arg parser object.
     */
    initializeDefaultArgParser(serviceMutation) {
        this._argumentParser = new argparse_1.ArgumentParser();
        this._argumentParser.addArgument(['-c', '--configuration'], {
            help: 'Override service configuration path',
        });
        if (serviceMutation) {
            serviceMutation(this);
        }
    }
    /**
     * Determine if backtrace-service should use backtrace-node integration
     */
    isBacktraceIntegrationEnabled() {
        return this.configuration.backtrace?.enabled === true;
    }
    /**
     * Generate correct credentials for metricsStorage obejct based on configuration file.
     */
    getMetricsSecret() {
        const isDiagnosticsEnabled = this.configuration.metrics?.enabled === true;
        if (isDiagnosticsEnabled === false) {
            return {
                uuid: '',
                secret: '',
            };
        }
        return {
            uuid: this.configuration.metrics?.uuid,
            secret: this.configuration.metrics?.secret,
        };
    }
}
exports.BacktraceService = BacktraceService;
//# sourceMappingURL=backtraceService.js.map